from abc import (ABC,
                 abstractmethod)
from itertools import groupby
from operator import attrgetter
from typing import (Iterable,
                    List,
                    Sequence,
                    Union as Union_)

from ground.base import (Context,
                         Orientation,
                         Relation)
from ground.hints import (Empty,
                          Mix,
                          Multipoint,
                          Multisegment,
                          Point,
                          Segment)
from reprit.base import generate_repr

from . import bounding
from .event import (LeftBinaryEvent as LeftEvent,
                    RightBinaryEvent as RightEvent)
from .events_queue import (LinearEventsQueue as BinaryEventsQueue,
                           NaryEventsQueue)
from .hints import SegmentEndpoints
from .operands import LinearOperand
from .sweep_line import (BinarySweepLine,
                         NarySweepLine)
from .unpacking import (unpack_linear_mix,
                        unpack_points,
                        unpack_segments)
from .utils import (all_equal,
                    endpoints_to_segments,
                    segments_to_endpoints,
                    to_endpoints,
                    to_segments_x_max)

Event = Union_[LeftEvent, RightEvent]


def merge_segments(segments: Sequence[Segment],
                   context: Context) -> Union_[Empty, Segment, Multisegment]:
    return unpack_segments(_merge_segments(segments, context), context)


def _merge_segments(segments: Sequence[Segment],
                    context: Context) -> Sequence[Segment]:
    events_queue = NaryEventsQueue(context)
    events_queue.register(segments_to_endpoints(segments))
    sweep_line = NarySweepLine(context)
    segments_endpoints = []  # type: List[SegmentEndpoints]
    while events_queue:
        event = events_queue.pop()
        if event.is_left:
            if event not in sweep_line:
                sweep_line.add(event)
                above_event, below_event = (sweep_line.above(event),
                                            sweep_line.below(event))
                if above_event is not None:
                    events_queue.detect_intersection(event, above_event)
                if below_event is not None:
                    events_queue.detect_intersection(below_event, event)
        else:
            event = event.left
            if event in sweep_line:
                above_event, below_event = (sweep_line.above(event),
                                            sweep_line.below(event))
                sweep_line.remove(event)
                if above_event is not None and below_event is not None:
                    events_queue.detect_intersection(below_event, above_event)
                segments_endpoints.append(to_endpoints(event))
    return endpoints_to_segments(
            sorted(endpoints for endpoints, _ in groupby(segments_endpoints)),
            context)


class Operation(ABC):
    __slots__ = 'context', 'first', 'second', '_events_queue'

    def __init__(self,
                 first: LinearOperand,
                 second: LinearOperand,
                 context: Context) -> None:
        """
        Initializes operation.

        :param first: first operand.
        :param second: second operand.
        :param context: operation context.
        """
        self.context, self.first, self.second = context, first, second
        self._events_queue = BinaryEventsQueue(context)

    __repr__ = generate_repr(__init__)

    @abstractmethod
    def compute(self) -> Union_[Empty, Mix, Multipoint, Multisegment, Segment]:
        """
        Computes result of the operation.
        """

    def fill_queue(self) -> None:
        self._events_queue.register(
                segments_to_endpoints(self.first.segments), True
        )
        self._events_queue.register(
                segments_to_endpoints(self.second.segments), False
        )

    def process_event(self,
                      event: Event,
                      sweep_line: BinarySweepLine) -> None:
        if event.is_left:
            sweep_line.add(event)
            above_event, below_event = (sweep_line.above(event),
                                        sweep_line.below(event))
            if above_event is not None:
                self._events_queue.detect_intersection(event, above_event)
            if below_event is not None:
                self._events_queue.detect_intersection(below_event, event)
        else:
            event = event.left
            if event in sweep_line:
                above_event, below_event = (sweep_line.above(event),
                                            sweep_line.below(event))
                sweep_line.remove(event)
                if above_event is not None and below_event is not None:
                    self._events_queue.detect_intersection(below_event,
                                                           above_event)

    def sweep(self) -> Iterable[LeftEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = BinarySweepLine(self.context)
        while events_queue:
            event = events_queue.pop()
            self.process_event(event, sweep_line)
            if not event.is_left:
                result.append(event.left)
        return result


class Difference(Operation):
    __slots__ = ()

    def compute(self) -> Union_[Empty, Segment, Multisegment]:
        context = self.context
        first_box, second_box = (context.segments_box(self.first.segments),
                                 context.segments_box(self.second.segments))
        if bounding.disjoint_with(first_box, second_box):
            return self.first.value
        self.second.segments = bounding.to_coupled_segments(
                first_box, self.second.segments, context
        )
        return (
            unpack_segments(
                    endpoints_to_segments(
                            sorted(endpoints
                                   for endpoints, events
                                   in groupby(self.sweep(),
                                              key=to_endpoints)
                                   if all(event.from_first_operand
                                          for event in events)),
                            context
                    ),
                    context
            )
            if self.second.segments
            else self.first.value
        )

    def sweep(self) -> Iterable[LeftEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = BinarySweepLine(self.context)
        first_x_max = to_segments_x_max(self.first.segments)
        while events_queue:
            event = events_queue.pop()
            if first_x_max < event.start.x:
                break
            self.process_event(event, sweep_line)
            if not event.is_left:
                result.append(event.left)
        return result


class Intersection(Operation):
    __slots__ = ()

    def compute(self) -> Union_[Empty, Segment, Multisegment]:
        context = self.context
        first_box, second_box = (context.segments_box(self.first.segments),
                                 context.segments_box(self.second.segments))
        if bounding.disjoint_with(first_box, second_box):
            return context.empty
        self.first.segments = bounding.to_coupled_segments(
                second_box, self.first.segments, context)
        if not self.first.segments:
            return context.empty
        self.second.segments = bounding.to_coupled_segments(
                first_box, self.second.segments, context)
        return (unpack_segments(endpoints_to_segments(
                sorted(endpoints
                       for endpoints, events in groupby(self.sweep(),
                                                        key=to_endpoints)
                       if not all_equal(event.from_first_operand
                                        for event in events)),
                context),
                context)
                if self.second.segments
                else context.empty)

    def sweep(self) -> Sequence[LeftEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = BinarySweepLine(self.context)
        min_max_x = min(to_segments_x_max(self.first.segments),
                        to_segments_x_max(self.second.segments))
        while events_queue:
            event = events_queue.pop()
            if min_max_x < event.start.x:
                break
            self.process_event(event, sweep_line)
            if not event.is_left:
                result.append(event.left)
        return result


class CompleteIntersection(Operation):
    __slots__ = ()

    def compute(self) -> Union_[Empty, Mix, Multipoint, Multisegment, Segment]:
        context = self.context
        first_box, second_box = (context.segments_box(self.first.segments),
                                 context.segments_box(self.second.segments))
        if bounding.disjoint_with(first_box, second_box):
            return context.empty
        self.first.segments = bounding.to_intersecting_segments(
                second_box, self.first.segments, context)
        if not self.first.segments:
            return context.empty
        self.second.segments = bounding.to_intersecting_segments(
                first_box, self.second.segments, context)
        if not self.second.segments:
            return context.empty
        points = []  # type: List[Point]
        endpoints = []  # type: List[SegmentEndpoints]
        for start, same_start_events_group in groupby(sorted(self.sweep(),
                                                             key=to_endpoints),
                                                      key=attrgetter('start')):
            same_start_events = list(same_start_events_group)
            if not all_equal(event.from_first_operand
                             for event in same_start_events):
                no_segment_found = True
                for event, next_event in zip(same_start_events,
                                             same_start_events[1:]):
                    if ((event.from_first_operand
                         is not next_event.from_first_operand)
                            and event.start == next_event.start
                            and event.end == next_event.end):
                        no_segment_found = False
                        if event.is_left:
                            endpoints.append(to_endpoints(event))
                if no_segment_found:
                    points.append(start)
        segments = endpoints_to_segments(endpoints, context)
        return unpack_linear_mix(unpack_points(points, context),
                                 unpack_segments(segments, context),
                                 context)

    def sweep(self) -> Sequence[LeftEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = BinarySweepLine(self.context)
        min_max_x = min(to_segments_x_max(self.first.segments),
                        to_segments_x_max(self.second.segments))
        while events_queue:
            event = events_queue.pop()
            if min_max_x < event.start.x:
                break
            self.process_event(event, sweep_line)
            result.append(event)
        return result


class SymmetricDifference(Operation):
    __slots__ = ()

    def compute(self) -> Union_[Empty, Segment, Multisegment]:
        context = self.context
        if bounding.disjoint_with(context.segments_box(self.first.segments),
                                  context.segments_box(self.second.segments)):
            segments = []
            segments += self.first.segments
            segments += self.second.segments
            segments.sort(key=to_endpoints)
            return context.multisegment_cls(segments)
        return unpack_segments(endpoints_to_segments(
                sorted(endpoints
                       for endpoints, events in groupby(self.sweep(),
                                                        key=to_endpoints)
                       if all_equal(event.from_first_operand
                                    for event in events)),
                context),
                context)


class Union(Operation):
    __slots__ = ()

    def compute(self) -> Multisegment:
        context = self.context
        if bounding.disjoint_with(context.segments_box(self.first.segments),
                                  context.segments_box(self.second.segments)):
            segments = []
            segments += self.first.segments
            segments += self.second.segments
            segments.sort(key=to_endpoints)
            return context.multisegment_cls(segments)
        return context.multisegment_cls(endpoints_to_segments(
                sorted(endpoints
                       for endpoints, _ in groupby(self.sweep(),
                                                   key=to_endpoints)),
                context))


def intersect_segments(first: Segment,
                       second: Segment,
                       context: Context) -> Union_[Empty, Multipoint, Segment]:
    relation = context.segments_relation(first, second)
    if relation is Relation.DISJOINT:
        return context.empty
    elif relation is Relation.TOUCH or relation is Relation.CROSS:
        return context.multipoint_cls([context.segments_intersection(first,
                                                                     second)])
    else:
        return (first
                if relation is Relation.EQUAL or relation is Relation.COMPONENT
                else
                (second
                 if relation is Relation.COMPOSITE
                 else _intersect_segments_overlap(first, second, context)))


def subtract_segments(minuend: Segment,
                      subtrahend: Segment,
                      context: Context
                      ) -> Union_[Empty, Multisegment, Segment]:
    relation = context.segments_relation(minuend, subtrahend)
    if relation is Relation.COMPONENT or relation is Relation.EQUAL:
        return context.empty
    elif (relation is Relation.DISJOINT
          or relation is Relation.TOUCH
          or relation is Relation.CROSS):
        return minuend
    else:
        return (_subtract_segments_overlap(minuend, subtrahend, context)
                if relation is Relation.OVERLAP
                else _subtract_segments_composite(minuend, subtrahend,
                                                  context))


def symmetric_subtract_segments(first: Segment,
                                second: Segment,
                                context: Context
                                ) -> Union_[Empty, Multisegment, Segment]:
    relation = context.segments_relation(first, second)
    if relation is Relation.DISJOINT:
        return context.multisegment_cls([first, second])
    elif relation is Relation.EQUAL:
        return context.empty
    else:
        return (_unite_segments_touch(first, second, context)
                if relation is Relation.TOUCH
                else
                (_unite_segments_cross(first, second, context)
                 if relation is Relation.CROSS
                 else
                 (_symmetric_subtract_segments_overlap(first, second, context)
                  if relation is Relation.OVERLAP
                  else (_subtract_segments_composite(first, second, context)
                        if relation is Relation.COMPOSITE
                        else _subtract_segments_composite(second, first,
                                                          context)))))


def unite_segments(first: Segment,
                   second: Segment,
                   context: Context) -> Union_[Multisegment, Segment]:
    relation = context.segments_relation(first, second)
    if relation is Relation.DISJOINT:
        return context.multisegment_cls([first, second])
    elif relation is Relation.EQUAL or relation is Relation.COMPOSITE:
        return first
    elif relation is Relation.COMPONENT:
        return second
    else:
        return (_unite_segments_touch
                if relation is Relation.TOUCH
                else (_unite_segments_cross
                      if relation is Relation.CROSS
                      else _unite_segments_overlap))(first, second, context)


def intersect_segment_with_multisegment(segment: Segment,
                                        multisegment: Multisegment,
                                        context: Context
                                        ) -> Union_[Empty, Multisegment,
                                                    Segment]:
    segments = _intersect_segment_with_multisegment(segment, multisegment,
                                                    context)
    return unpack_segments(segments, context)


def subtract_segment_from_multisegment(multisegment: Multisegment,
                                       segment: Segment,
                                       context: Context
                                       ) -> Union_[Empty, Multisegment,
                                                   Segment]:
    segments = _subtract_segment_from_multisegment(multisegment, segment,
                                                   context)
    return unpack_segments(segments, context)


def unite_segment_with_multisegment(segment: Segment,
                                    multisegment: Multisegment,
                                    context: Context
                                    ) -> Union_[Multisegment, Segment]:
    segments = _subtract_segment_from_multisegment(multisegment, segment,
                                                   context)
    segments.append(segment)
    return unpack_segments(segments, context)


def _intersect_segment_with_multisegment(segment: Segment,
                                         multisegment: Multisegment,
                                         context: Context) -> List[Segment]:
    result = []
    for sub_segment in multisegment.segments:
        relation = context.segments_relation(segment, sub_segment)
        if (relation is not Relation.DISJOINT
                and relation is not Relation.TOUCH
                and relation is not Relation.CROSS):
            result.append(segment
                          if (relation is Relation.EQUAL
                              or relation is Relation.COMPONENT)
                          else (sub_segment
                                if relation is Relation.COMPOSITE
                                else _intersect_segments_overlap(segment,
                                                                 sub_segment,
                                                                 context)))
    return result


def _subtract_segment_from_multisegment(multisegment: Multisegment,
                                        segment: Segment,
                                        context: Context) -> List[Segment]:
    result = []
    segment_cls = context.segment_cls
    for index, sub_segment in enumerate(multisegment.segments):
        relation = context.segments_relation(segment, sub_segment)
        if relation is Relation.EQUAL:
            result.extend(multisegment.segments[index + 1:])
            break
        elif relation is Relation.COMPONENT:
            left_start, left_end, right_start, right_end = sorted([
                sub_segment.start, sub_segment.end, segment.start,
                segment.end])
            result.extend(
                    [segment_cls(right_start, right_end)]
                    if left_start == segment.start or left_start == segment.end
                    else
                    ((([segment_cls(left_start, left_end)]
                       if right_start == right_end
                       else [segment_cls(left_start, left_end),
                             segment_cls(right_start, right_end)])
                      if (right_start == segment.start
                          or right_start == segment.end)
                      else [segment_cls(left_start, left_end)])
                     if left_end == segment.start or left_end == segment.end
                     else [segment_cls(left_start, right_start)]))
            result.extend(multisegment.segments[index + 1:])
            break
        elif relation is Relation.CROSS:
            cross_point = context.segments_intersection(sub_segment, segment)
            result.append(segment_cls(sub_segment.start, cross_point))
            result.append(segment_cls(cross_point, sub_segment.end))
        elif relation is Relation.OVERLAP:
            result.append(_subtract_segments_overlap(sub_segment, segment,
                                                     context))
        elif relation is not Relation.COMPOSITE:
            result.append(sub_segment)
    return result


def _subtract_segments_composite(minuend: Segment,
                                 subtrahend: Segment,
                                 context: Context
                                 ) -> Union_[Multisegment, Segment]:
    left_start, left_end, right_start, right_end = sorted([
        minuend.start, minuend.end, subtrahend.start, subtrahend.end])
    return (context.segment_cls(right_start, right_end)
            if left_start == subtrahend.start or left_start == subtrahend.end
            else (((context.segment_cls(left_start, left_end)
                    if right_start == right_end
                    else
                    context.multisegment_cls([context.segment_cls(left_start,
                                                                  left_end),
                                              context.segment_cls(right_start,
                                                                  right_end)]))
                   if (right_start == subtrahend.start
                       or right_start == subtrahend.end)
                   else context.segment_cls(left_start, left_end))
                  if left_end == subtrahend.start or left_end == subtrahend.end
                  else context.segment_cls(left_start, right_start)))


def _subtract_segments_overlap(minuend: Segment,
                               subtrahend: Segment,
                               context: Context) -> Segment:
    left_start, left_end, right_start, right_end = sorted([
        minuend.start, minuend.end, subtrahend.start, subtrahend.end])
    return (context.segment_cls(left_start, left_end)
            if left_start == minuend.start or left_start == minuend.end
            else context.segment_cls(right_start, right_end))


def _symmetric_subtract_segments_overlap(minuend: Segment,
                                         subtrahend: Segment,
                                         context: Context) -> Multisegment:
    left_start, left_end, right_start, right_end = sorted([
        minuend.start, minuend.end, subtrahend.start, subtrahend.end])
    return context.multisegment_cls([context.segment_cls(left_start, left_end),
                                     context.segment_cls(right_start,
                                                         right_end)])


def _unite_segments_cross(first: Segment,
                          second: Segment,
                          context: Context) -> Multisegment:
    cross_point = context.segments_intersection(first, second)
    segment_cls = context.segment_cls
    return context.multisegment_cls([segment_cls(first.start, cross_point),
                                     segment_cls(second.start, cross_point),
                                     segment_cls(cross_point, first.end),
                                     segment_cls(cross_point, second.end)])


def _unite_segments_overlap(first: Segment,
                            second: Segment,
                            context: Context) -> Segment:
    start, _, _, end = sorted([first.start, first.end, second.start,
                               second.end])
    return context.segment_cls(start, end)


def _unite_segments_touch(first: Segment,
                          second: Segment,
                          context: Context) -> Union_[Multisegment, Segment]:
    return (context.multisegment_cls([first, second])
            if ((first.start != second.start
                 or (context.angle_orientation(first.start, first.end,
                                               second.end)
                     is not Orientation.COLLINEAR))
                and (first.start != second.end
                     or (context.angle_orientation(first.start, first.end,
                                                   second.start)
                         is not Orientation.COLLINEAR))
                and (first.end != second.start
                     or (context.angle_orientation(first.end, first.start,
                                                   second.end)
                         is not Orientation.COLLINEAR))
                and (first.end != second.end
                     or (context.angle_orientation(first.end, first.start,
                                                   second.start)
                         is not Orientation.COLLINEAR)))
            else context.segment_cls(min(first.start, first.end,
                                         second.start, second.end),
                                     max(first.start, first.end,
                                         second.start, second.end)))


def _intersect_segments_overlap(first: Segment,
                                second: Segment,
                                context: Context) -> Segment:
    _, start, end, _ = sorted([first.start, first.end, second.start,
                               second.end])
    return context.segment_cls(start, end)



from abc import (ABC,
                 abstractmethod)
from itertools import groupby
from operator import attrgetter
from typing import (Iterable,
                    List,
                    Optional,
                    Union as Union_)

from ground.base import Context
from ground.hints import (Empty,
                          Mix,
                          Multipoint,
                          Multipolygon,
                          Multisegment,
                          Point,
                          Polygon,
                          Segment)
from reprit.base import generate_repr

from . import bounding
from .event import (LeftMixedEvent as LeftEvent,
                    RightMixedEvent as RightEvent)
from .events_queue import MixedEventsQueue as EventsQueue
from .operands import (HoleyOperand,
                       LinearOperand)
from .sweep_line import BinarySweepLine as SweepLine
from .unpacking import (unpack_linear_mix,
                        unpack_points,
                        unpack_segments)
from .utils import (all_equal,
                    endpoints_to_segments,
                    polygon_to_oriented_edges_endpoints,
                    segments_to_endpoints,
                    to_endpoints,
                    to_polygons_x_max,
                    to_segments_x_max)

Event = Union_[LeftEvent, RightEvent]


class Operation(ABC):
    __slots__ = 'context', 'linear', 'shaped', '_events_queue'

    def __init__(self,
                 linear: LinearOperand,
                 shaped: HoleyOperand,
                 context: Context) -> None:
        """
        Initializes operation.

        :param linear: first operand.
        :param shaped: second operand.
        :param context: operation context.
        """
        self.context, self.linear, self.shaped = context, linear, shaped
        self._events_queue = EventsQueue(context)

    __repr__ = generate_repr(__init__)

    @abstractmethod
    def compute(self) -> Union_[Empty, Mix, Multipoint, Multisegment]:
        """
        Computes result of the operation.
        """

    def compute_fields(self,
                       event: LeftEvent,
                       below_event: Optional[LeftEvent]) -> None:
        if below_event is not None:
            event.other_interior_to_left = (
                below_event.other_interior_to_left
                if event.from_first_operand is below_event.from_first_operand
                else below_event.interior_to_left
            )
        event.from_result = self.from_result(event)

    def fill_queue(self) -> None:
        events_queue = self._events_queue
        events_queue.register(segments_to_endpoints(self.linear.segments),
                              True)
        for polygon in self.shaped.polygons:
            events_queue.register(
                    polygon_to_oriented_edges_endpoints(polygon, self.context),
                    False
            )

    @abstractmethod
    def from_result(self, event: LeftEvent) -> bool:
        """Detects if event is a part of resulting geometry."""

    def process_event(self, event: Event, sweep_line: SweepLine) -> None:
        if not event.is_left:
            event = event.left
            if event in sweep_line:
                above_event, below_event = (sweep_line.above(event),
                                            sweep_line.below(event))
                sweep_line.remove(event)
                if above_event is not None and below_event is not None:
                    self._events_queue.detect_intersection(below_event,
                                                           above_event)
        elif event not in sweep_line:
            sweep_line.add(event)
            above_event, below_event = (sweep_line.above(event),
                                        sweep_line.below(event))
            self.compute_fields(event, below_event)
            if (above_event is not None
                    and self._events_queue.detect_intersection(event,
                                                               above_event)):
                self.compute_fields(event, below_event)
                self.compute_fields(above_event, event)
            if (below_event is not None
                    and self._events_queue.detect_intersection(below_event,
                                                               event)):
                below_below_event = sweep_line.below(below_event)
                self.compute_fields(below_event, below_below_event)
                self.compute_fields(event, below_event)


class Difference(Operation):
    __slots__ = ()

    def compute(self) -> Union_[Empty, Multisegment, Segment]:
        context = self.context
        segments_box = context.segments_box(self.linear.segments)
        if bounding.disjoint_with(segments_box,
                                  context.polygons_box(self.shaped.polygons)):
            return self.linear.value
        self.shaped.polygons = bounding.to_coupled_polygons(
                segments_box, self.shaped.polygons, context)
        if not self.shaped.polygons:
            return self.linear.value
        return unpack_segments(endpoints_to_segments([to_endpoints(event)
                                                      for event in self.sweep()
                                                      if event.from_result],
                                                     context),
                               context)

    def from_result(self, event: LeftEvent) -> bool:
        return event.from_first_operand and event.outside

    def sweep(self) -> Iterable[LeftEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = SweepLine(self.context)
        first_x_max = to_segments_x_max(self.linear.segments)
        while events_queue:
            event = events_queue.pop()
            if first_x_max < event.start.x:
                break
            self.process_event(event, sweep_line)
            if not event.is_left:
                result.append(event.left)
        return result


class CompleteIntersection(Operation):
    __slots__ = ()

    def compute(self) -> Union_[Empty, Mix, Multipoint, Multisegment, Segment]:
        context = self.context
        linear_box, shaped_box = (context.segments_box(self.linear.segments),
                                  context.polygons_box(self.shaped.polygons))
        if bounding.disjoint_with(linear_box, shaped_box):
            return context.empty
        self.linear.segments = bounding.to_intersecting_segments(
                shaped_box, self.linear.segments, context)
        if not self.linear.segments:
            return context.empty
        self.shaped.polygons = bounding.to_intersecting_polygons(
                linear_box, self.shaped.polygons, context)
        if not self.shaped.polygons:
            return context.empty
        events = sorted(self.sweep(),
                        key=self._events_queue.key)
        points = []  # type: List[Point]
        for start, same_start_events in groupby(events,
                                                key=attrgetter('start')):
            same_start_events = list(same_start_events)
            if not (any(event.primary.from_result
                        for event in same_start_events)
                    or all_equal(event.from_first_operand
                                 for event in same_start_events)):
                points.append(start)
        segments = endpoints_to_segments(
                [to_endpoints(event)
                 for event in events
                 if event.is_left and event.from_result],
                context)
        return unpack_linear_mix(unpack_points(points, context),
                                 unpack_segments(segments, context),
                                 context)

    def from_result(self, event: LeftEvent) -> bool:
        return event.from_first_operand and not event.outside

    def sweep(self) -> Iterable[LeftEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = SweepLine(self.context)
        min_max_x = min(to_segments_x_max(self.linear.segments),
                        to_polygons_x_max(self.shaped.polygons))
        while events_queue:
            event = events_queue.pop()
            if min_max_x < event.start.x:
                break
            self.process_event(event, sweep_line)
            result.append(event)
        return result


class Intersection(Operation):
    __slots__ = ()

    def compute(self) -> Union_[Empty, Segment, Multisegment]:
        context = self.context
        linear_box, shaped_box = (context.segments_box(self.linear.segments),
                                  context.polygons_box(self.shaped.polygons))
        if bounding.disjoint_with(linear_box, shaped_box):
            return context.empty
        self.linear.segments = bounding.to_intersecting_segments(
                shaped_box, self.linear.segments, context)
        if not self.linear.segments:
            return context.empty
        self.shaped.polygons = bounding.to_intersecting_polygons(
                linear_box, self.shaped.polygons, context)
        if not self.shaped.polygons:
            return context.empty
        segments = endpoints_to_segments([to_endpoints(event)
                                          for event in self.sweep()
                                          if event.from_result],
                                         context)
        return unpack_segments(segments, context)

    def from_result(self, event: LeftEvent) -> bool:
        return event.from_first_operand and not event.outside

    def sweep(self) -> Iterable[LeftEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = SweepLine(self.context)
        min_max_x = min(to_segments_x_max(self.linear.segments),
                        to_polygons_x_max(self.shaped.polygons))
        while events_queue:
            event = events_queue.pop()
            if min_max_x < event.start.x:
                break
            self.process_event(event, sweep_line)
            if event.is_left:
                result.append(event)
        return result


class Union(Operation):
    __slots__ = ()

    def compute(self) -> Union_[Empty, Mix, Multipolygon, Multisegment,
                                Polygon, Segment]:
        context = self.context
        linear_box, shaped_box = (context.segments_box(self.linear.segments),
                                  context.polygons_box(self.shaped.polygons))
        if bounding.disjoint_with(linear_box, shaped_box):
            return context.mix_cls(context.empty, self.linear.value,
                                   self.shaped.value)
        result_segments, self.linear.segments = (
            bounding.split_intersecting_segments(
                    shaped_box, self.linear.segments, context))
        self.shaped.polygons = bounding.to_intersecting_polygons(
                linear_box, self.shaped.polygons, context)
        segments = endpoints_to_segments([to_endpoints(event)
                                          for event in self.sweep()
                                          if event.from_result], context)
        segments.extend(result_segments)
        linear = unpack_segments(segments, context)
        return (self.shaped.value
                if linear is context.empty
                else context.mix_cls(context.empty, linear, self.shaped.value))

    def from_result(self, event: LeftEvent) -> bool:
        return event.from_first_operand and event.outside

    def sweep(self) -> Iterable[LeftEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = SweepLine(self.context)
        while events_queue:
            event = events_queue.pop()
            self.process_event(event, sweep_line)
            if event.is_left:
                result.append(event)
        return result


SymmetricDifference = Union



from abc import (ABC,
                 abstractmethod)
from typing import Union

from ground.hints import (Multipolygon,
                          Multisegment,
                          Polygon,
                          Segment)
from reprit.base import generate_repr

from .hints import (Multiregion,
                    Region)


class LinearOperand(ABC):
    __slots__ = 'segments',

    @property
    @abstractmethod
    def value(self) -> Union[Multisegment, Segment]:
        """Returns value of the operand."""


class MultisegmentOperand(LinearOperand):
    __slots__ = '_value'

    def __init__(self, value: Multisegment) -> None:
        self.segments, self._value = value.segments, value

    @property
    def value(self) -> Multisegment:
        return self._value


class SegmentOperand(LinearOperand):
    __slots__ = '_value'

    def __init__(self, value: Segment) -> None:
        self.segments, self._value = [value], value

    @property
    def value(self) -> Segment:
        return self._value


class HolelessOperand(ABC):
    __slots__ = 'regions',

    @property
    @abstractmethod
    def value(self) -> Union[Multiregion, Region]:
        """Returns value of the operand."""


class MultiregionOperand(HolelessOperand):
    __slots__ = '_value',

    def __init__(self, value: Multiregion) -> None:
        self._value, self.regions = value, value

    __repr__ = generate_repr(__init__)

    @property
    def value(self) -> Multiregion:
        return self._value


class RegionOperand(HolelessOperand):
    __slots__ = '_value',

    def __init__(self, value: Region) -> None:
        self._value, self.regions = value, [value]

    __repr__ = generate_repr(__init__)

    @property
    def value(self) -> Region:
        return self._value


class HoleyOperand(ABC):
    __slots__ = 'polygons',

    @property
    @abstractmethod
    def value(self) -> Union[Multipolygon, Polygon]:
        """Returns value of the operand."""


class MultipolygonOperand(HoleyOperand):
    __slots__ = '_value',

    def __init__(self, value: Multipolygon) -> None:
        self._value, self.polygons = value, value.polygons

    __repr__ = generate_repr(__init__)

    @property
    def value(self) -> Multipolygon:
        return self._value


class PolygonOperand(HoleyOperand):
    __slots__ = '_value',

    def __init__(self, value: Polygon) -> None:
        self._value, self.polygons = value, [value]

    __repr__ = generate_repr(__init__)

    @property
    def value(self) -> Polygon:
        return self._value



from itertools import (chain,
                       groupby)
from typing import (Any,
                    Iterable,
                    List,
                    Sequence)

from ground.base import (Context,
                         Orientation)
from ground.hints import (Contour,
                          Point,
                          Polygon,
                          Scalar,
                          Segment)

from .hints import (Orienteer,
                    Region,
                    SegmentEndpoints)


def all_equal(iterable: Iterable[Any]) -> bool:
    groups = groupby(iterable)
    return next(groups, True) and not next(groups, False)


def polygon_to_oriented_edges_endpoints(polygon: Polygon,
                                        context: Context
                                        ) -> Iterable[SegmentEndpoints]:
    yield from contour_to_oriented_edges_endpoints(polygon.border, context)
    for hole in polygon.holes:
        yield from contour_to_oriented_edges_endpoints(hole, context, True)


def contour_to_oriented_edges_endpoints(contour: Contour,
                                        context: Context,
                                        clockwise: bool = False
                                        ) -> Iterable[SegmentEndpoints]:
    vertices = contour.vertices
    return (((vertices[index - 1], vertices[index])
             for index in range(len(vertices)))
            if (to_contour_orientation(contour, context)
                is (Orientation.CLOCKWISE
                    if clockwise
                    else Orientation.COUNTERCLOCKWISE))
            else ((vertices[index], vertices[index - 1])
                  for index in range(len(vertices) - 1, -1, -1)))


def to_contour_orientation(contour: Contour, context: Context) -> Orientation:
    vertices = contour.vertices
    index = min(range(len(vertices)),
                key=vertices.__getitem__)
    return context.angle_orientation(vertices[index - 1], vertices[index],
                                     vertices[(index + 1) % len(vertices)])


flatten = chain.from_iterable


def to_first_border_vertex(polygon: Polygon) -> Point:
    return polygon.border.vertices[0]


def to_regions_x_max(regions: Sequence[Region]) -> Scalar:
    return max(vertex.x
               for border in regions
               for vertex in border.vertices)


def to_polygons_x_max(polygons: Sequence[Polygon]) -> Scalar:
    return max(vertex.x
               for polygon in polygons
               for vertex in polygon.border.vertices)


def to_segments_x_max(segments: Sequence[Segment]) -> Scalar:
    return max(max(segment.start.x, segment.end.x) for segment in segments)


def shrink_collinear_vertices(vertices: List[Point],
                              orienteer: Orienteer) -> None:
    index = -len(vertices) + 1
    while index < 0:
        while (max(2, -index) < len(vertices)
               and orienteer(vertices[index + 1], vertices[index + 2],
                             vertices[index]) is Orientation.COLLINEAR):
            del vertices[index + 1]
        index += 1
    while index < len(vertices):
        while (max(2, index) < len(vertices)
               and orienteer(vertices[index - 1], vertices[index - 2],
                             vertices[index]) is Orientation.COLLINEAR):
            del vertices[index - 1]
        index += 1


def segments_to_endpoints(segments: Sequence[Segment]
                          ) -> Iterable[SegmentEndpoints]:
    return (to_endpoints(segment) for segment in segments)


def endpoints_to_segments(endpoints: Iterable[SegmentEndpoints],
                          context: Context) -> List[Segment]:
    segment_cls = context.segment_cls
    return [segment_cls(start, end) for start, end in endpoints]


def to_endpoints(segment: Segment) -> SegmentEndpoints:
    return segment.start, segment.end



from functools import partial
from typing import (Callable,
                    Generic,
                    Iterable,
                    Type,
                    Union,
                    cast)

from ground.base import (Context,
                         Orientation,
                         Relation)
from ground.hints import Point
from prioq.base import PriorityQueue
from reprit.base import generate_repr

from .enums import OverlapKind
from .event import (LeftBinaryEvent,
                    LeftHolelessEvent,
                    LeftHoleyEvent,
                    LeftMixedEvent,
                    LeftNaryEvent,
                    LeftShapedEvent,
                    RightBinaryEvent,
                    RightMixedEvent,
                    RightNaryEvent,
                    RightShapedEvent)
from .hints import (Orienteer,
                    SegmentEndpoints)

BinaryEvent = Union[LeftBinaryEvent, RightBinaryEvent]
NaryEvent = Union[LeftNaryEvent, RightNaryEvent]
MixedEvent = Union[LeftMixedEvent, RightMixedEvent]
ShapedEvent = Union[LeftShapedEvent, RightShapedEvent]


class BinaryEventsQueueKey:
    __slots__ = 'event', 'orienteer'

    def __init__(self, orienteer: Orienteer, event: BinaryEvent) -> None:
        self.orienteer, self.event = orienteer, event

    __repr__ = generate_repr(__init__)

    def __lt__(self, other: 'BinaryEventsQueueKey') -> bool:
        event, other_event = self.event, other.event
        start, other_start = event.start, other_event.start
        if start.x != other_start.x:
            # different x-coordinate,
            # the event with lower x-coordinate is processed first
            return start.x < other_start.x
        elif start.y != other_start.y:
            # different points, but same x-coordinate,
            # the event with lower y-coordinate is processed first
            return start.y < other_start.y
        elif event.is_left is not other_event.is_left:
            # same start, but one is a left endpoint
            # and the other a right endpoint,
            # the right endpoint is processed first
            return not event.is_left
        # same start, both events are left endpoints
        # or both are right endpoints
        else:
            other_end_orientation = self.orienteer(event.start, event.end,
                                                   other_event.end)
            # the lowest segment is processed first
            return (other_event.from_first_operand
                    if other_end_orientation is Orientation.COLLINEAR
                    else (other_end_orientation
                          # the lowest segment is processed first
                          is (Orientation.COUNTERCLOCKWISE
                              if event.is_left
                              else Orientation.CLOCKWISE)))


class NaryEventsQueueKey:
    __slots__ = 'event', 'orienteer'

    def __init__(self, orienteer: Orienteer, event: NaryEvent) -> None:
        self.orienteer, self.event = orienteer, event

    __repr__ = generate_repr(__init__)

    def __lt__(self, other: 'NaryEventsQueueKey') -> bool:
        event, other_event = self.event, other.event
        start, other_start = event.start, other_event.start
        if start.x != other_start.x:
            # different x-coordinate,
            # the event with lower x-coordinate is processed first
            return start.x < other_start.x
        elif start.y != other_start.y:
            # different points, but same x-coordinate,
            # the event with lower y-coordinate is processed first
            return start.y < other_start.y
        elif event.is_left is not other_event.is_left:
            # same start, but one is a left endpoint
            # and the other a right endpoint,
            # the right endpoint is processed first
            return not event.is_left
        # same start, both events are left endpoints
        # or both are right endpoints
        else:
            # the lowest segment is processed first
            return (self.orienteer(event.start, event.end, other_event.end)
                    is (Orientation.COUNTERCLOCKWISE
                        if event.is_left
                        else Orientation.CLOCKWISE))


class LinearEventsQueue:
    __slots__ = 'context', '_queue'

    def __init__(self, context: Context) -> None:
        self.context = context
        self._queue = PriorityQueue(key=partial(BinaryEventsQueueKey,
                                                context.angle_orientation))

    __repr__ = generate_repr(__init__)

    def __bool__(self) -> bool:
        return bool(self._queue)

    @property
    def key(self) -> Callable[[BinaryEvent], BinaryEventsQueueKey]:
        return self._queue.key

    def detect_intersection(self,
                            below_event: LeftBinaryEvent,
                            event: LeftBinaryEvent) -> None:
        relation = self.context.segments_relation(below_event, event)
        if relation is Relation.CROSS or relation is Relation.TOUCH:
            if (event.start != below_event.start
                    and event.end != below_event.end):
                # segments do not intersect at endpoints
                point = self.context.segments_intersection(below_event, event)
                if point != below_event.start and point != below_event.end:
                    self._divide_segment(below_event, point)
                if point != event.start and point != event.end:
                    self._divide_segment(event, point)
        elif relation is not Relation.DISJOINT:
            # segments overlap
            if below_event.from_first_operand is event.from_first_operand:
                raise ValueError('Segments of the same multisegment '
                                 'should not overlap.')
            starts_equal = below_event.start == event.start
            if starts_equal:
                start_min = start_max = None
            elif self.key(event) < self.key(below_event):
                start_min, start_max = event, below_event
            else:
                start_min, start_max = below_event, event
            ends_equal = event.end == below_event.end
            if ends_equal:
                end_min = end_max = None
            elif self.key(event.opposite) < self.key(below_event.opposite):
                end_min, end_max = event.opposite, below_event.opposite
            else:
                end_min, end_max = below_event.opposite, event.opposite
            if starts_equal:
                # both line segments are equal or share the left endpoint
                if not ends_equal:
                    self._divide_segment(end_max.opposite, end_min.start)
            elif ends_equal:
                # the line segments share the right endpoint
                self._divide_segment(start_min, start_max.start)
            elif start_min is end_max.opposite:
                # one line segment includes the other one
                self._divide_segment(start_min, end_min.start)
                self._divide_segment(start_min, start_max.start)
            else:
                # no line segment includes the other one
                self._divide_segment(start_max, end_min.start)
                self._divide_segment(start_min, start_max.start)

    def pop(self) -> BinaryEvent:
        return self._queue.pop()

    def register(self,
                 segments_endpoints: Iterable[SegmentEndpoints],
                 from_first_operand: bool) -> None:
        events_queue = self._queue
        for segment_endpoints in segments_endpoints:
            event = LeftBinaryEvent.from_segment_endpoints(segment_endpoints,
                                                           from_first_operand)
            events_queue.push(event)
            events_queue.push(event.opposite)

    def _divide_segment(self, event: LeftBinaryEvent, point: Point) -> None:
        tail = event.divide(point)
        self._queue.push(tail)
        self._queue.push(event.opposite)


class MixedEventsQueue:
    __slots__ = 'context', '_queue'

    def __init__(self, context: Context) -> None:
        self.context = context
        self._queue = PriorityQueue(key=partial(BinaryEventsQueueKey,
                                                context.angle_orientation))

    @property
    def key(self) -> Callable[[MixedEvent], BinaryEventsQueueKey]:
        return self._queue.key

    __repr__ = generate_repr(__init__)

    def __bool__(self) -> bool:
        return bool(self._queue)

    def detect_intersection(self,
                            below_event: LeftMixedEvent,
                            event: LeftMixedEvent) -> bool:
        relation = self.context.segments_relation(below_event, event)
        if relation is Relation.CROSS or relation is Relation.TOUCH:
            if (event.start != below_event.start
                    and event.end != below_event.end):
                # segments do not intersect at endpoints
                point = self.context.segments_intersection(below_event, event)
                if point != below_event.start and point != below_event.end:
                    self._divide_segment(below_event, point)
                if point != event.start and point != event.end:
                    self._divide_segment(event, point)
        elif relation is not Relation.DISJOINT:
            # segments overlap
            if below_event.from_first_operand is event.from_first_operand:
                raise ValueError('Edges of the {geometry} '
                                 'should not overlap.'
                                 .format(geometry=('multisegment'
                                                   if event.from_first_operand
                                                   else 'multipolygon')))
            event.is_overlap = below_event.is_overlap = True
            starts_equal = below_event.start == event.start
            if starts_equal:
                start_min = start_max = None
            elif self.key(event) < self.key(below_event):
                start_min, start_max = event, below_event
            else:
                start_min, start_max = below_event, event
            ends_equal = event.end == below_event.end
            if ends_equal:
                end_min = end_max = None
            elif self.key(event.opposite) < self.key(below_event.opposite):
                end_min, end_max = event.opposite, below_event.opposite
            else:
                end_min, end_max = below_event.opposite, event.opposite
            if starts_equal:
                # both line segments are equal or share the left endpoint
                if not ends_equal:
                    self._divide_segment(end_max.opposite, end_min.start)
                return True
            elif ends_equal:
                # the line segments share the right endpoint
                self._divide_segment(start_min, start_max.start)
            elif start_min is end_max.opposite:
                # one line segment includes the other one
                self._divide_segment(start_min, end_min.start)
                self._divide_segment(start_min, start_max.start)
            else:
                # no line segment includes the other one
                self._divide_segment(start_max, end_min.start)
                self._divide_segment(start_min, start_max.start)
        return False

    def pop(self) -> MixedEvent:
        return self._queue.pop()

    def register(self,
                 segments_endpoints: Iterable[SegmentEndpoints],
                 from_first_operand: bool) -> None:
        push = self._queue.push
        for segment_endpoints in segments_endpoints:
            event = LeftMixedEvent.from_endpoints(segment_endpoints,
                                                  from_first_operand)
            push(event)
            push(event.opposite)

    def _divide_segment(self, event: LeftMixedEvent, point: Point) -> None:
        tail = event.divide(point)
        self._queue.push(tail)
        self._queue.push(event.opposite)


class NaryEventsQueue:
    __slots__ = 'context', '_queue'

    def __init__(self, context: Context) -> None:
        self.context = context
        self._queue = PriorityQueue(key=partial(NaryEventsQueueKey,
                                                context.angle_orientation))

    __repr__ = generate_repr(__init__)

    def __bool__(self) -> bool:
        return bool(self._queue)

    @property
    def key(self) -> Callable[[NaryEvent], NaryEventsQueueKey]:
        return self._queue.key

    def detect_intersection(self,
                            below_event: LeftNaryEvent,
                            event: LeftNaryEvent) -> None:
        relation = self.context.segments_relation(below_event, event)
        if relation is Relation.CROSS or relation is Relation.TOUCH:
            if (event.start != below_event.start
                    and event.end != below_event.end):
                # segments do not intersect at endpoints
                point = self.context.segments_intersection(below_event, event)
                if point != below_event.start and point != below_event.end:
                    self._divide_segment(below_event, point)
                if point != event.start and point != event.end:
                    self._divide_segment(event, point)
        elif relation is not Relation.DISJOINT:
            # segments overlap
            starts_equal = below_event.start == event.start
            if starts_equal:
                start_min = start_max = None
            elif self.key(event) < self.key(below_event):
                start_min, start_max = event, below_event
            else:
                start_min, start_max = below_event, event
            ends_equal = event.end == below_event.end
            if ends_equal:
                end_min = end_max = None
            elif self.key(event.opposite) < self.key(below_event.opposite):
                end_min, end_max = event.opposite, below_event.opposite
            else:
                end_min, end_max = below_event.opposite, event.opposite
            if starts_equal:
                # both line segments are equal or share the left endpoint
                if not ends_equal:
                    self._divide_segment(end_max.opposite, end_min.start)
            elif ends_equal:
                # the line segments share the right endpoint
                self._divide_segment(start_min, start_max.start)
            elif start_min is end_max.opposite:
                # one line segment includes the other one
                self._divide_segment(start_min, end_min.start)
                self._divide_segment(start_min, start_max.start)
            else:
                # no line segment includes the other one
                self._divide_segment(start_max, end_min.start)
                self._divide_segment(start_min, start_max.start)

    def pop(self) -> NaryEvent:
        return self._queue.pop()

    def register(self, segments_endpoints: Iterable[SegmentEndpoints]) -> None:
        push = self._queue.push
        for segment_endpoints in segments_endpoints:
            event = LeftNaryEvent.from_segment_endpoints(segment_endpoints)
            push(event)
            push(event.opposite)

    def _divide_segment(self, event: LeftNaryEvent, point: Point) -> None:
        tail = LeftNaryEvent.divide(event, point)
        self._queue.push(tail)
        self._queue.push(event.opposite)


class ShapedEventsQueue(Generic[LeftShapedEvent]):
    __slots__ = 'context', 'event_cls', '_queue'

    def __init__(self,
                 event_cls: Type[LeftShapedEvent],
                 context: Context) -> None:
        self.event_cls, self.context = event_cls, context
        self._queue = PriorityQueue(key=partial(BinaryEventsQueueKey,
                                                context.angle_orientation))

    __repr__ = generate_repr(__init__)

    def __bool__(self) -> bool:
        return bool(self._queue)

    @property
    def key(self) -> Callable[[ShapedEvent], BinaryEventsQueueKey]:
        return self._queue.key

    def detect_intersection(self,
                            below_event: LeftShapedEvent,
                            event: LeftShapedEvent) -> bool:
        relation = self.context.segments_relation(below_event, event)
        if relation is Relation.CROSS or relation is Relation.TOUCH:
            if (event.start != below_event.start
                    and event.end != below_event.end):
                # segments do not intersect at endpoints
                point = self.context.segments_intersection(below_event, event)
                if point != below_event.start and point != below_event.end:
                    self._divide_segment(below_event, point)
                if point != event.start and point != event.end:
                    self._divide_segment(event, point)
        elif relation is not Relation.DISJOINT:
            # segments overlap
            if below_event.from_first_operand is event.from_first_operand:
                raise ValueError('Edges of the same geometry '
                                 'should not overlap.')
            starts_equal = below_event.start == event.start
            if starts_equal:
                start_min = start_max = None
            elif self.key(event) < self.key(below_event):
                start_min, start_max = event, below_event
            else:
                start_min, start_max = below_event, event
            ends_equal = event.end == below_event.end
            if ends_equal:
                end_min = end_max = None
            elif self.key(event.opposite) < self.key(below_event.opposite):
                end_min, end_max = event.opposite, below_event.opposite
            else:
                end_min, end_max = below_event.opposite, event.opposite
            if starts_equal:
                # both line segments are equal or share the left endpoint
                below_event.overlap_kind = event.overlap_kind = (
                    OverlapKind.SAME_ORIENTATION
                    if event.interior_to_left is below_event.interior_to_left
                    else OverlapKind.DIFFERENT_ORIENTATION
                )
                if not ends_equal:
                    self._divide_segment(end_max.opposite, end_min.start)
                return True
            elif ends_equal:
                # the line segments share the right endpoint
                self._divide_segment(start_min, start_max.start)
            elif start_min is end_max.opposite:
                # one line segment includes the other one
                self._divide_segment(start_min, end_min.start)
                self._divide_segment(start_min, start_max.start)
            else:
                # no line segment includes the other one
                self._divide_segment(start_max, end_min.start)
                self._divide_segment(start_min, start_max.start)
        return False

    def pop(self) -> ShapedEvent:
        return self._queue.pop()

    def register(self,
                 segments_endpoints: Iterable[SegmentEndpoints],
                 from_first_operand: bool) -> None:
        event_cls, push = self.event_cls, self._queue.push
        for segment_endpoints in segments_endpoints:
            event = event_cls.from_endpoints(segment_endpoints,
                                             from_first_operand)
            push(event)
            push(event.opposite)

    def _divide_segment(self, event: LeftShapedEvent, point: Point) -> None:
        tail = event.divide(point)
        self._queue.push(tail)
        self._queue.push(event.opposite)


HolelessEventsQueue = cast(Callable[[Context],
                                    ShapedEventsQueue[LeftHolelessEvent]],
                           partial(ShapedEventsQueue, LeftHolelessEvent))
HoleyEventsQueue = cast(Callable[[Context],
                                 ShapedEventsQueue[LeftHoleyEvent]],
                        partial(ShapedEventsQueue, LeftHoleyEvent))



import sys
from abc import (ABC,
                 abstractmethod)
from reprlib import recursive_repr
from typing import (Optional,
                    Sequence,
                    TypeVar)

from ground.hints import Point
from reprit import seekers
from reprit.base import generate_repr

from .enums import OverlapKind
from .hints import SegmentEndpoints

UNDEFINED_INDEX = sys.maxsize


class Event(ABC):
    start: Point

    __slots__ = ()

    @property
    def end(self) -> Point:
        return self.opposite.start

    @property
    @abstractmethod
    def is_left(self) -> bool:
        """Checks if the event's start corresponds to the leftmost endpoint."""

    @property
    @abstractmethod
    def opposite(self) -> 'Event':
        """Returns opposite of the event."""


class LeftEvent(Event):
    is_left = True

    __slots__ = 'right',

    @property
    def opposite(self) -> 'RightEvent':
        return self.right

    @abstractmethod
    def divide(self, point: Point) -> 'Event':
        """Divides the event at given break point and returns tail."""


class RightEvent(Event):
    is_left = False

    @property
    def opposite(self) -> LeftEvent:
        return self.left

    __slots__ = 'left',


class LeftNaryEvent(LeftEvent):
    @classmethod
    def from_segment_endpoints(
            cls, segment_endpoints: SegmentEndpoints
    ) -> 'LeftNaryEvent':
        start, end = segment_endpoints
        if start > end:
            start, end = end, start
        result = cls(start, None)
        result.right = RightNaryEvent(end, result)
        return result

    __slots__ = '_start',

    def __init__(self,
                 start: Point,
                 right: Optional['RightNaryEvent']) -> None:
        self.right, self._start = right, start

    __repr__ = recursive_repr()(generate_repr(__init__))

    @property
    def start(self) -> Point:
        return self._start

    def divide(self, break_point: Point) -> 'LeftNaryEvent':
        tail = self.right.left = LeftNaryEvent(break_point, self.right)
        self.right = RightNaryEvent(break_point, self)
        return tail


class RightNaryEvent(RightEvent):
    __slots__ = 'start',

    def __init__(self, start: Point, left: LeftNaryEvent) -> None:
        self.left, self.start = left, start

    __repr__ = recursive_repr()(generate_repr(__init__))


class LeftBinaryEvent(LeftEvent):
    @classmethod
    def from_segment_endpoints(cls,
                               segment_endpoints: SegmentEndpoints,
                               from_first_operand: bool) -> 'LeftBinaryEvent':
        start, end = segment_endpoints
        if start > end:
            start, end = end, start
        event = cls(start, None, from_first_operand)
        event.right = RightBinaryEvent(end, event)
        return event

    __slots__ = 'from_first_operand', 'start'

    def __init__(self,
                 start: Point,
                 right: Optional['RightBinaryEvent'],
                 from_first_operand: bool) -> None:
        self.from_first_operand, self.right, self.start = (
            from_first_operand, right, start
        )

    __repr__ = recursive_repr()(generate_repr(__init__))

    def divide(self, point: Point) -> 'LeftBinaryEvent':
        tail = self.right.left = LeftBinaryEvent(point, self.right,
                                                 self.from_first_operand)
        self.right = RightBinaryEvent(point, self)
        return tail


class RightBinaryEvent(RightEvent):
    __slots__ = 'start',

    def __init__(self, start: Point, left: LeftBinaryEvent) -> None:
        self.left, self.start = left, start

    __repr__ = recursive_repr()(generate_repr(__init__))

    @property
    def from_first_operand(self) -> bool:
        return self.left.from_first_operand


class LeftMixedEvent(LeftEvent):
    @classmethod
    def from_endpoints(cls,
                       segment_endpoints: SegmentEndpoints,
                       from_first_operand: bool) -> 'LeftMixedEvent':
        start, end = segment_endpoints
        inside_on_left = True
        if start > end:
            start, end = end, start
            inside_on_left = False
        result = cls(start, None, from_first_operand, inside_on_left)
        result.right = RightMixedEvent(end, result)
        return result

    __slots__ = ('from_first_operand', 'from_result', 'interior_to_left',
                 'is_overlap', 'other_interior_to_left', 'start')

    def __init__(self,
                 start: Point,
                 right: Optional['RightMixedEvent'],
                 from_first_operand: bool,
                 interior_to_left: bool,
                 other_interior_to_left: bool = False,
                 is_overlap: bool = False,
                 from_result: bool = False) -> None:
        self.right, self.start = right, start
        self.from_first_operand = from_first_operand
        self.interior_to_left, self.other_interior_to_left = (
            interior_to_left, other_interior_to_left
        )
        self.is_overlap = is_overlap
        self.from_result = from_result

    __repr__ = recursive_repr()(generate_repr(__init__))

    @property
    def outside(self) -> bool:
        """
        Checks if the segment touches or disjoint with the intersection.
        """
        return not self.other_interior_to_left and not self.is_overlap

    @property
    def primary(self) -> 'LeftMixedEvent':
        return self

    def divide(self, point: Point) -> 'LeftMixedEvent':
        tail = self.right.left = LeftMixedEvent(
                point, self.right, self.from_first_operand,
                self.interior_to_left
        )
        self.right = RightMixedEvent(point, self)
        return tail


class RightMixedEvent(RightEvent):
    __slots__ = 'start',

    def __init__(self, start: Point, left: LeftMixedEvent) -> None:
        self.left, self.start = left, start

    __repr__ = recursive_repr()(generate_repr(__init__))

    @property
    def from_first_operand(self) -> bool:
        return self.left.from_first_operand

    @property
    def primary(self) -> LeftMixedEvent:
        return self.left


class LeftHolelessEvent(LeftEvent):
    @classmethod
    def from_endpoints(cls,
                       segment_endpoints: SegmentEndpoints,
                       from_first_operand: bool) -> 'LeftHolelessEvent':
        start, end = segment_endpoints
        inside_on_left = True
        if start > end:
            start, end = end, start
            inside_on_left = False
        event = cls(start, None, from_first_operand, inside_on_left)
        event.right = RightShapedEvent(end, event)
        return event

    __slots__ = ('from_first_operand', 'from_shaped_result', 'id',
                 'interior_to_left', 'other_interior_to_left', 'overlap_kind',
                 'start')

    def __init__(self,
                 start: Point,
                 right: Optional['RightShapedEvent'],
                 from_first_operand: bool,
                 interior_to_left: bool,
                 *,
                 from_shaped_result: bool = False,
                 id_: int = 0,
                 other_interior_to_left: bool = False,
                 overlap_kind: OverlapKind = OverlapKind.NONE) -> None:
        (
            self.from_first_operand, self.from_shaped_result, self.id,
            self.interior_to_left, self.other_interior_to_left,
            self.overlap_kind, self.right, self.start
        ) = (
            from_first_operand, from_shaped_result, id_, interior_to_left,
            other_interior_to_left, overlap_kind, right, start
        )

    __repr__ = recursive_repr()(generate_repr(__init__,
                                              field_seeker=seekers.complex_))

    @property
    def inside(self) -> bool:
        """
        Checks if the segment enclosed by
        or lies within the region of the intersection.
        """
        return (self.other_interior_to_left
                and self.overlap_kind is OverlapKind.NONE)

    @property
    def is_common_polyline_component(self) -> bool:
        """
        Checks if the segment is a component of intersection's polyline.
        """
        return self.overlap_kind is OverlapKind.DIFFERENT_ORIENTATION

    @property
    def is_common_region_boundary(self) -> bool:
        """
        Checks if the segment is a boundary of intersection's region.
        """
        return self.overlap_kind is OverlapKind.SAME_ORIENTATION

    @property
    def is_overlap(self) -> bool:
        """
        Checks if the segment lies on the boundary of both operands.
        """
        return self.overlap_kind is not OverlapKind.NONE

    @property
    def outside(self) -> bool:
        """
        Checks if the segment touches or disjoint with the intersection.
        """
        return (not self.other_interior_to_left
                and self.overlap_kind is OverlapKind.NONE)

    @property
    def primary(self) -> 'LeftHolelessEvent':
        return self

    @property
    def wholly_in_complete_intersection(self) -> bool:
        """
        Checks if the segment wholly is a part of the complete intersection.
        """
        return self.from_shaped_result or self.is_common_polyline_component

    def divide(self, point: Point) -> 'LeftHolelessEvent':
        tail = self.right.left = LeftHolelessEvent(
                point, self.right, self.from_first_operand,
                self.interior_to_left
        )
        self.right = RightShapedEvent(point, self)
        return tail


class LeftHoleyEvent(LeftEvent):
    @classmethod
    def from_endpoints(cls,
                       segment_endpoints: SegmentEndpoints,
                       from_first_operand: bool) -> 'LeftHoleyEvent':
        start, end = segment_endpoints
        interior_to_left = True
        if start > end:
            start, end = end, start
            interior_to_left = False
        event = cls(start, None, from_first_operand, interior_to_left)
        event.right = RightShapedEvent(end, event)
        return event

    __slots__ = ('below_event_from_shaped_result', 'contour_id',
                 'from_first_operand', 'from_in_to_out', 'from_shaped_result',
                 'interior_to_left', 'other_interior_to_left', 'overlap_kind',
                 'id', 'start', 'start_id')

    def __init__(
            self,
            start: Point,
            right: Optional['RightShapedEvent'],
            from_first_operand: bool,
            interior_to_left: bool,
            *,
            below_event_from_shaped_result: Optional['LeftHoleyEvent'] = None,
            contour_id: Optional[int] = None,
            from_in_to_out: bool = False,
            from_shaped_result: bool = False,
            id_: int = UNDEFINED_INDEX,
            other_interior_to_left: bool = False,
            overlap_kind: OverlapKind = OverlapKind.NONE,
            start_id: int = UNDEFINED_INDEX
    ) -> None:
        (
            self.below_event_from_shaped_result, self.contour_id,
            self.from_first_operand, self.from_in_to_out,
            self.from_shaped_result, self.id, self.interior_to_left,
            self.other_interior_to_left, self.overlap_kind, self.right,
            self.start, self.start_id
        ) = (
            below_event_from_shaped_result, contour_id, from_first_operand,
            from_in_to_out, from_shaped_result, id_, interior_to_left,
            other_interior_to_left, overlap_kind, right, start, start_id
        )

    __repr__ = recursive_repr()(generate_repr(__init__,
                                              field_seeker=seekers.complex_))

    @property
    def end_id(self) -> int:
        return self.opposite.start_id

    @property
    def inside(self) -> bool:
        """
        Checks if the segment enclosed by
        or lies within the region of the intersection.
        """
        return (self.other_interior_to_left
                and self.overlap_kind is OverlapKind.NONE)

    @property
    def is_common_polyline_component(self) -> bool:
        """
        Checks if the segment is a component of intersection's polyline.
        """
        return self.overlap_kind is OverlapKind.DIFFERENT_ORIENTATION

    @property
    def is_common_region_boundary(self) -> bool:
        """
        Checks if the segment is a boundary of intersection's region.
        """
        return self.overlap_kind is OverlapKind.SAME_ORIENTATION

    @property
    def is_overlap(self) -> bool:
        """
        Checks if the segment lies on the boundary of both operands.
        """
        return self.overlap_kind is not OverlapKind.NONE

    @property
    def is_vertical(self) -> bool:
        return self.start.x == self.end.x

    @property
    def primary(self) -> 'LeftHoleyEvent':
        return self

    @property
    def outside(self) -> bool:
        """
        Checks if the segment touches or disjoint with the intersection.
        """
        return (not self.other_interior_to_left
                and self.overlap_kind is OverlapKind.NONE)

    @property
    def wholly_in_complete_intersection(self) -> bool:
        """
        Checks if the segment wholly is a part of the complete intersection.
        """
        return self.from_shaped_result or self.is_common_polyline_component

    def divide(self, point: Point) -> 'LeftHoleyEvent':
        tail = self.right.left = LeftHoleyEvent(
                point, self.right, self.from_first_operand,
                self.interior_to_left
        )
        self.right = RightShapedEvent(point, self)
        return tail


LeftShapedEvent = TypeVar('LeftShapedEvent', LeftHolelessEvent, LeftHoleyEvent)


class RightShapedEvent(RightEvent):
    __slots__ = 'id', 'start', 'start_id'

    def __init__(self,
                 start: Point,
                 left: LeftShapedEvent,
                 *,
                 id_: int = UNDEFINED_INDEX,
                 start_id: int = UNDEFINED_INDEX) -> None:
        self.id, self.left, self.start, self.start_id = (
            id_, left, start, start_id
        )

    __repr__ = recursive_repr()(generate_repr(__init__,
                                              field_seeker=seekers.complex_))

    @property
    def end_id(self) -> int:
        return self.opposite.start_id

    @property
    def from_first_operand(self) -> bool:
        return self.left.from_first_operand

    @property
    def primary(self) -> LeftShapedEvent:
        return self.left


def events_to_connectivity(events: Sequence[Event]) -> Sequence[int]:
    events_count = len(events)
    result = [0] * events_count
    index = 0
    while index < events_count:
        current_start = events[index].start
        right_start_index = index
        while (index < events_count
               and events[index].start == current_start
               and not events[index].is_left):
            index += 1
        left_start_index = index
        while index < events_count and events[index].start == current_start:
            index += 1
        left_stop_index = index - 1
        has_right_events = left_start_index >= right_start_index + 1
        has_left_events = left_stop_index >= left_start_index
        if has_right_events:
            result[right_start_index:left_start_index - 1] = range(
                    right_start_index + 1, left_start_index - 1 + 1
            )
            result[left_start_index - 1] = (left_stop_index
                                            if has_left_events
                                            else right_start_index)
        if has_left_events:
            result[left_start_index] = (right_start_index
                                        if has_right_events
                                        else left_stop_index)
            result[left_start_index + 1:left_stop_index + 1] = range(
                    left_start_index, left_stop_index
            )
    return result



from typing import (Sequence,
                    Tuple)

from ground.base import (Context,
                         Location,
                         Relation)
from ground.hints import (Box,
                          Contour,
                          Point,
                          Polygon,
                          Segment)
from orient.planar import (point_in_region,
                           segment_in_contour,
                           segment_in_region)

from .hints import (Multiregion,
                    Region)


def disjoint_with(left: Box, right: Box) -> bool:
    """
    Checks if boxes do not intersect.

    >>> from ground.base import get_context
    >>> context = get_context()
    >>> Box = context.box_cls
    >>> disjoint_with(Box(0, 2, 0, 2), Box(0, 2, 0, 2))
    False
    >>> disjoint_with(Box(0, 2, 0, 2), Box(1, 3, 1, 3))
    False
    >>> disjoint_with(Box(0, 2, 0, 2), Box(2, 4, 0, 2))
    False
    >>> disjoint_with(Box(0, 2, 0, 2), Box(2, 4, 2, 4))
    False
    >>> disjoint_with(Box(0, 2, 0, 2), Box(2, 4, 3, 5))
    True
    """
    return (right.max_x < left.min_x or left.max_x < right.min_x
            or right.max_y < left.min_y or left.max_y < right.min_y)


def intersects_with(left: Box, right: Box) -> bool:
    """
    Checks if boxes intersect.

    >>> from ground.base import get_context
    >>> context = get_context()
    >>> Box = context.box_cls
    >>> intersects_with(Box(0, 2, 0, 2), Box(0, 2, 0, 2))
    True
    >>> intersects_with(Box(0, 2, 0, 2), Box(1, 3, 1, 3))
    True
    >>> intersects_with(Box(0, 2, 0, 2), Box(2, 4, 0, 2))
    True
    >>> intersects_with(Box(0, 2, 0, 2), Box(2, 4, 2, 4))
    True
    >>> intersects_with(Box(0, 2, 0, 2), Box(2, 4, 3, 5))
    False
    """
    return (right.min_x <= left.max_x and left.min_x <= right.max_x
            and right.min_y <= left.max_y and left.min_y <= right.max_y)


def coupled_with(left: Box, right: Box) -> bool:
    """
    Checks if boxes intersect in some region or by the edge.

    >>> from ground.base import get_context
    >>> context = get_context()
    >>> Box = context.box_cls
    >>> coupled_with(Box(0, 2, 0, 2), Box(0, 2, 0, 2))
    True
    >>> coupled_with(Box(0, 2, 0, 2), Box(1, 3, 1, 3))
    True
    >>> coupled_with(Box(0, 2, 0, 2), Box(2, 4, 0, 2))
    True
    >>> coupled_with(Box(0, 2, 0, 2), Box(2, 4, 2, 4))
    False
    >>> coupled_with(Box(0, 2, 0, 2), Box(2, 4, 3, 5))
    False
    """
    return (intersects_with(left, right)
            and (not touches_with(left, right)
                 or edges_overlap_with(left, right)))


def touches_with(left: Box, right: Box) -> bool:
    """
    Checks if boxes intersect at point or by the edge.

    >>> from ground.base import get_context
    >>> context = get_context()
    >>> Box = context.box_cls
    >>> touches_with(Box(0, 2, 0, 2), Box(0, 2, 0, 2))
    False
    >>> touches_with(Box(0, 2, 0, 2), Box(1, 3, 1, 3))
    False
    >>> touches_with(Box(0, 2, 0, 2), Box(2, 4, 0, 2))
    True
    >>> touches_with(Box(0, 2, 0, 2), Box(2, 4, 2, 4))
    True
    >>> touches_with(Box(0, 2, 0, 2), Box(2, 4, 3, 5))
    False
    """
    return ((left.min_x == right.max_x or left.max_x == right.min_x)
            and (left.min_y <= right.max_y and right.min_y <= left.max_y)
            or (left.min_x <= right.max_x and right.min_x <= left.max_x)
            and (left.min_y == right.max_y or right.min_y == left.max_y))


def edges_overlap_with(left: Box, right: Box) -> bool:
    """
    Checks if boxes intersect by the edge.

    >>> from ground.base import get_context
    >>> context = get_context()
    >>> Box = context.box_cls
    >>> edges_overlap_with(Box(0, 2, 0, 2), Box(0, 2, 0, 2))
    False
    >>> edges_overlap_with(Box(0, 2, 0, 2), Box(1, 3, 1, 3))
    False
    >>> edges_overlap_with(Box(0, 2, 0, 2), Box(2, 4, 0, 2))
    True
    >>> edges_overlap_with(Box(0, 2, 0, 2), Box(2, 4, 2, 4))
    False
    >>> edges_overlap_with(Box(0, 2, 0, 2), Box(2, 4, 3, 5))
    False
    """
    return ((left.min_x == right.max_x or left.max_x == right.min_x)
            and (left.min_y < right.max_y and right.min_y < left.max_y)
            or (left.min_x < right.max_x and right.min_x < left.max_x)
            and (left.min_y == right.max_y or right.min_y == left.max_y))


def is_subset_of(test: Box, goal: Box) -> bool:
    """
    Checks if the box is the subset of the other.

    >>> from ground.base import get_context
    >>> context = get_context()
    >>> Box = context.box_cls
    >>> is_subset_of(Box(0, 2, 0, 2), Box(0, 2, 0, 2))
    True
    >>> is_subset_of(Box(0, 2, 0, 2), Box(1, 3, 1, 3))
    False
    >>> is_subset_of(Box(0, 2, 0, 2), Box(2, 4, 0, 2))
    False
    >>> is_subset_of(Box(0, 2, 0, 2), Box(2, 4, 2, 4))
    False
    >>> is_subset_of(Box(0, 2, 0, 2), Box(2, 4, 3, 5))
    False
    """
    return (goal.min_x <= test.min_x and test.max_x <= goal.max_x
            and goal.min_y <= test.min_y and test.max_y <= goal.max_y)


def within_of(test: Box, goal: Box) -> bool:
    """
    Checks if the box is contained in an interior of the other.

    >>> from ground.base import get_context
    >>> context = get_context()
    >>> Box = context.box_cls
    >>> within_of(Box(0, 2, 0, 2), Box(0, 2, 0, 2))
    False
    >>> within_of(Box(0, 2, 0, 2), Box(1, 3, 1, 3))
    False
    >>> within_of(Box(0, 2, 0, 2), Box(2, 4, 0, 2))
    False
    >>> within_of(Box(0, 2, 0, 2), Box(2, 4, 2, 4))
    False
    >>> within_of(Box(0, 2, 0, 2), Box(2, 4, 3, 5))
    False
    """
    return (goal.min_x < test.min_x and test.max_x < goal.max_x
            and goal.min_y < test.min_y and test.max_y < goal.max_y)


def intersects_with_segment(box: Box,
                            segment: Segment,
                            context: Context) -> bool:
    """
    Checks if the box intersects the segment.
    """
    segment_box = context.segment_box(segment)
    return (intersects_with(segment_box, box)
            and (is_subset_of(segment_box, box)
                 or any(context.segments_relation(edge, segment)
                        is not Relation.DISJOINT
                        for edge in to_edges(box, context))))


def coupled_with_segment(box: Box,
                         segment: Segment,
                         context: Context) -> bool:
    """
    Checks if the box intersects the segment at more than one point.
    """
    segment_box = context.segment_box(segment)
    return (coupled_with(segment_box, box)
            and (is_subset_of(segment_box, box)
                 or any(context.segments_relation(edge, segment)
                        not in (Relation.TOUCH, Relation.DISJOINT)
                        for edge in to_edges(box, context))))


def is_subset_of_region(box: Box,
                        border: Contour,
                        context: Context) -> bool:
    """
    Checks if the box is the subset of the region.
    """
    return all(segment_in_region(segment, border) in (Relation.COMPONENT,
                                                      Relation.ENCLOSED,
                                                      Relation.WITHIN)
               for segment in to_edges(box, context))


def within_of_region(box: Box,
                     border: Contour,
                     context: Context) -> bool:
    """
    Checks if the box is contained in an interior of the region.
    """
    return (all(point_in_region(vertex, border) is Location.INTERIOR
                for vertex in to_vertices(box, context))
            and all(context.segments_relation(edge, border_edge)
                    is Relation.DISJOINT
                    for edge in to_edges(box, context)
                    for border_edge in context.contour_segments(border)))


def is_subset_of_multiregion(box: Box,
                             multiregion: Multiregion,
                             context: Context) -> bool:
    """
    Checks if the box is the subset of the multiregion.
    """
    return any(is_subset_of(box, context.contour_box(region))
               and is_subset_of_region(box, region, context)
               for region in multiregion)


def intersects_with_polygon(box: Box,
                            polygon: Polygon,
                            context: Context) -> bool:
    """
    Checks if the box intersects the polygon.
    """
    border = polygon.border
    polygon_box = context.contour_box(border)
    if not intersects_with(polygon_box, box):
        return False
    elif (is_subset_of(polygon_box, box)
          or any(contains_point(box, vertex) for vertex in border.vertices)):
        return True
    locations = [point_in_region(vertex, border)
                 for vertex in to_vertices(box, context)]
    if (within_of(box, polygon_box)
            and all(location is Location.INTERIOR for location in locations)
            and all(context.segments_relation(edge, border_edge)
                    is Relation.DISJOINT
                    for edge in to_edges(box, context)
                    for border_edge in context.contour_segments(border))):
        return not any(within_of(box, context.contour_box(hole))
                       and within_of_region(box, hole, context)
                       for hole in polygon.holes)
    else:
        return (any(location is not Location.EXTERIOR
                    for location in locations)
                or any(intersects_with_segment(box, edge, context)
                       for edge in context.contour_segments(border)))


def intersects_with_region(box: Box,
                           region: Region,
                           context: Context) -> bool:
    """
    Checks if the box intersects the region.
    """
    region_box = context.contour_box(region)
    return (intersects_with(region_box, box)
            and (is_subset_of(region_box, box)
                 or any(contains_point(box, vertex)
                        for vertex in region.vertices)
                 or any(point_in_region(vertex, region)
                        is not Location.EXTERIOR
                        for vertex in to_vertices(box, context))
                 or any(intersects_with_segment(box, edge, context)
                        for edge in context.contour_segments(region))))


def coupled_with_polygon(box: Box,
                         polygon: Polygon,
                         context: Context) -> bool:
    """
    Checks if the box intersects the polygon in continuous points set.
    """
    border = polygon.border
    polygon_box = context.contour_box(border)
    if not coupled_with(polygon_box, box):
        return False
    elif (is_subset_of(polygon_box, box)
          or any(covers_point(box, vertex) for vertex in border.vertices)):
        return True
    locations = [point_in_region(vertex, border)
                 for vertex in to_vertices(box, context)]
    if any(location is Location.INTERIOR for location in locations):
        return (not all(location is Location.INTERIOR
                        for location in locations)
                or not is_subset_of_multiregion(box, polygon.holes, context))
    else:
        return (not is_subset_of_multiregion(box, polygon.holes, context)
                if (is_subset_of(box, polygon_box)
                    and is_subset_of_region(box, border, context))
                else any(segment_in_contour(segment, border)
                         is Relation.OVERLAP
                         or segment_in_region(segment, border)
                         in (Relation.CROSS, Relation.COMPONENT,
                             Relation.ENCLOSED)
                         for segment in to_edges(box, context)))


def coupled_with_region(box: Box,
                        region: Region,
                        context: Context) -> bool:
    """
    Checks if the box intersects the region in continuous points set.
    """
    region_box = context.contour_box(region)
    if not coupled_with(region_box, box):
        return False
    elif (is_subset_of(region_box, box)
          or any(covers_point(box, vertex) for vertex in region.vertices)):
        return True
    return (any(point_in_region(vertex, region) is Location.INTERIOR
                for vertex in to_vertices(box, context))
            or is_subset_of(box, region_box)
            and is_subset_of_region(box, region, context)
            or any(segment_in_contour(segment, region)
                   is Relation.OVERLAP
                   or segment_in_region(segment, region)
                   in (Relation.CROSS, Relation.COMPONENT,
                       Relation.ENCLOSED)
                   for segment in to_edges(box, context)))


def contains_point(box: Box, point: Point) -> bool:
    return (box.min_x <= point.x <= box.max_x
            and box.min_y <= point.y <= box.max_y)


def covers_point(box: Box, point: Point) -> bool:
    return box.min_x < point.x < box.max_x and box.min_y < point.y < box.max_y


def split_intersecting_segments(box: Box,
                                segments: Sequence[Segment],
                                context: Context) -> Tuple[Sequence[Segment],
                                                           Sequence[Segment]]:
    result = [], []
    for segment in segments:
        result[intersects_with_segment(box, segment, context)].append(segment)
    return result


def to_vertices(box: Box,
                context: Context) -> Sequence[Point]:
    point_cls = context.point_cls
    return (point_cls(box.min_x, box.min_y), point_cls(box.max_x, box.min_y),
            point_cls(box.max_x, box.max_y), point_cls(box.min_x, box.max_y))


def to_edges(box: Box,
             context: Context) -> Sequence[Segment]:
    point_cls, segment_cls = context.point_cls, context.segment_cls
    return (segment_cls(point_cls(box.min_x, box.min_y),
                        point_cls(box.max_x, box.min_y)),
            segment_cls(point_cls(box.max_x, box.min_y),
                        point_cls(box.max_x, box.max_y)),
            segment_cls(point_cls(box.min_x, box.max_y),
                        point_cls(box.max_x, box.max_y)),
            segment_cls(point_cls(box.min_x, box.min_y),
                        point_cls(box.min_x, box.max_y)))


def to_intersecting_segments(box: Box,
                             segments: Sequence[Segment],
                             context: Context) -> Sequence[Segment]:
    return [segment
            for segment in segments
            if intersects_with_segment(box, segment, context)]


def to_coupled_segments(box: Box,
                        segments: Sequence[Segment],
                        context: Context) -> Sequence[Segment]:
    return [segment
            for segment in segments
            if coupled_with_segment(box, segment, context)]


def to_intersecting_polygons(box: Box,
                             polygons: Sequence[Polygon],
                             context: Context) -> Sequence[Polygon]:
    return [polygon
            for polygon in polygons
            if intersects_with_polygon(box, polygon, context)]


def to_intersecting_regions(box: Box,
                            multiregion: Multiregion,
                            context: Context) -> Multiregion:
    return [region
            for region in multiregion
            if intersects_with_region(box, region, context)]


def to_coupled_polygons(box: Box,
                        polygons: Sequence[Polygon],
                        context: Context) -> Sequence[Polygon]:
    return [polygon
            for polygon in polygons
            if coupled_with_polygon(box, polygon, context)]


def to_coupled_regions(box: Box,
                       multiregion: Multiregion,
                       context: Context) -> Multiregion:
    return [region
            for region in multiregion
            if coupled_with_region(box, region, context)]



from typing import (Sequence,
                    Union)

from ground.base import Context
from ground.hints import (Contour,
                          Empty,
                          Linear,
                          Maybe,
                          Mix,
                          Multipoint,
                          Multipolygon,
                          Multisegment,
                          Point,
                          Polygon,
                          Segment,
                          Shaped)


def unpack_linear_mix(discrete: Maybe[Multipoint],
                      linear: Maybe[Linear],
                      context: Context
                      ) -> Union[Empty, Mix, Multipoint, Linear]:
    return (linear
            if discrete is context.empty
            else (discrete
                  if linear is context.empty
                  else context.mix_cls(discrete, linear, context.empty)))


def unpack_mix(discrete: Multipoint,
               linear: Linear,
               shaped: Maybe[Shaped],
               context: Context
               ) -> Union[Empty, Linear, Mix, Multipoint, Shaped]:
    empty = context.empty
    return ((shaped
             if linear is empty
             else (linear
                   if shaped is empty
                   else context.mix_cls(discrete, linear, shaped)))
            if discrete is empty
            else (discrete
                  if shaped is empty and linear is empty
                  else context.mix_cls(discrete, linear, shaped)))


def unpack_points(points: Sequence[Point], context: Context
                  ) -> Union[Empty, Multipoint]:
    return context.multipoint_cls(points) if points else context.empty


def unpack_polygons(polygons: Sequence[Polygon],
                    context: Context) -> Union[Empty, Multipolygon, Polygon]:
    return ((context.multipolygon_cls(polygons)
             if len(polygons) > 1
             else polygons[0])
            if polygons
            else context.empty)


def unpack_regions(regions: Sequence[Contour],
                   context: Context) -> Union[Empty, Polygon, Multipolygon]:
    polygon_cls = context.polygon_cls
    return ((context.multipolygon_cls([polygon_cls(region, [])
                                       for region in regions])
             if len(regions) > 1
             else polygon_cls(regions[0], []))
            if regions
            else context.empty)


def unpack_segments(segments: Sequence[Segment],
                    context: Context) -> Union[Empty, Multisegment, Segment]:
    return ((context.multisegment_cls(segments)
             if len(segments) > 1
             else segments[0])
            if segments
            else context.empty)
